import re

from six import string_types

from ksql.error import SQLTypeNotImplementYetError, \
					   IllegalTableTypeError, \
					   IllegalValueFormatError, \
					   SQLFormatNotImplementError, \
					   BuildNotImplmentError

class SQLBuilder(object):

	@classmethod
	def build(self, sql_type, **kwargs):
		if sql_type == 'create':
			sql_builder = CreateBuilder(kwargs.pop('table_type'))
			sql_str = sql_builder.build(kwargs.pop('table_name'), 
							  kwargs.pop('columns_type'),
							  kwargs.pop('topic'),
							  kwargs.pop('value_format'),
							  kwargs.pop('key', None))

		elif sql_type == 'create_as':
			sql_builder = CreateAsBuilder(kwargs.pop('table_type'))
			sql_str = sql_builder.build(table_name = kwargs.pop('table_name'), 
							  			select_columns = kwargs.pop('select_columns', None),
							  			src_table = kwargs.pop('src_table'),
							  			kafka_topic = kwargs.pop('kafka_topic'),
							  			value_format = kwargs.pop('value_format'),
							  			partition_by = kwargs.pop('partition_by', None),
							  			**kwargs)

		else:
			raise SQLTypeNotImplementYetError(sql_type)

		return sql_str


class BaseCreateBuilder(object):
	def __init__(self, table_type, sql_format=None):
		self.table_types = ['table', 'stream']
		self.value_formats = ['delimited', 'json', 'avro']
		self.table_type = table_type
		self.sql_format = sql_format

		if table_type.lower() not in self.table_types:
			raise IllegalTableTypeError(table_type)

		if not sql_format:
			raise SQLFormatNotImplementError()

	def build(self):
		raise BuildNotImplmentError()

	def _parsed_with_properties(self, **kwargs):
		properties = ['']
		for prop in self.properties:
			if prop in kwargs.keys():
				value = kwargs.get(prop)
				if isinstance(value, string_types):
					p = "{}='{}'".format(prop, kwargs.get(prop))
				else:
					p = "{}={}".format(prop, kwargs.get(prop))
				
				properties.append(p)

		return ', '.join(properties)


class CreateBuilder(BaseCreateBuilder):
	def __init__(self, table_type):
		str_format = "CREATE {} {} ({}) WITH (kafka_topic='{}', value_format='{}'{});"
		super(CreateBuilder, self).__init__(table_type, str_format)
		self.properties = ['kafka_topic', 'value_format', 'key', 'timestamp']


	def build(self, table_name, columns_type=[], topic=None, value_format='DELIMITED', key=None):
		if value_format.lower() not in self.value_formats:
			raise IllegalValueFormatError(value_format)
			
		built_colums_type = self._build_colums_type(columns_type)
		built_key = self._build_key(key)

		sql_str = self.sql_format.format(self.table_type, table_name, built_colums_type, topic, value_format, built_key)
		return sql_str

	@staticmethod
	def _build_colums_type(columns_type):
		built_columns_type = ', '.join(columns_type)
		return built_columns_type

	@staticmethod
	def _build_key(key):
		if key:
			built_key = ", key='{}'".format(key)
			return built_key
		else:
			return ''


class CreateAsBuilder(BaseCreateBuilder):
	def __init__(self, table_type):
		str_format = "CREATE {} {} WITH (kafka_topic='{}', value_format='{}'{}) AS SELECT {} FROM {} {} {}"
		super(CreateAsBuilder, self).__init__(table_type, str_format)
		self.properties = ['kafka_topic', 'value_format', 'partitions', 'replicas', 'timestamp']

		
	def build(self, table_name, select_columns, src_table, kafka_topic=None, 
			  value_format='DELIMITED', conditions=[], partition_by=None, **kwargs):

		if value_format.lower() not in self.value_formats:
			raise IllegalValueFormatError(value_format)
		
		if not kafka_topic:
			kafka_topic = table_name

		select_clause, where_clause, partition_by_clause, properties = \
			self._build_clauses(select_columns, conditions, partition_by, **kwargs)

		sql_str = self.sql_format.format(self.table_type, table_name, 
										kafka_topic, value_format, properties, 
										select_clause, src_table, where_clause, 
										partition_by_clause)

		cleaned_sql_str = re.sub( '\s+', ' ', sql_str).strip()

		return cleaned_sql_str

	def _build_clauses(self, select_columns, conditions, partition_by, **kwargs):
		select_clause = self._build_select_clause(select_columns)
		where_clause = self._build_where_clause(conditions)
		partition_by_clause = self._build_partition_by_clause(partition_by)
		properties = self._parsed_with_properties(**kwargs)

		return select_clause, where_clause, partition_by_clause, properties

	@staticmethod
	def _build_where_clause(conditions):
		if len(conditions) > 0:
			where_clause = "where {}".format(conditions).replace('"','\'')
			return where_clause
		else:
			return ''

	@staticmethod
	def _build_select_clause(select_columns):
		select_clause = '*'
		if select_columns:
			if len(select_columns) > 0:
				select_clause = ', '.join(select_columns)

		return select_clause

	@staticmethod
	def _build_partition_by_clause(partition_by):
		if partition_by:
			return 'PARTITION BY {}'.format(partition_by)
		else:
			return ''



import functools
import json
import time

import requests
from  requests import Timeout

from ksql.builder import SQLBuilder
from ksql.error import CreateError

class BaseAPI(object):
    def __init__(self, url, **kwargs):
        self.url = url
        self.max_retries = kwargs.get("max_retries", 3)
        self.delay = kwargs.get("delay", 0)
        self.timeout = kwargs.get("timeout", 5)

    def get_timout(self):
        return self.timeout

    @staticmethod
    def _validate_sql_string(sql_string):
        if len(sql_string) > 0:
            if sql_string[-1] != ';':
                sql_string += ';'
        return sql_string

    @staticmethod
    def _parse_ksql_res(r, error):
        if_current_status = r[0].get('currentStatus')
        if if_current_status:
            r = r[0]['currentStatus']['commandStatus']['status']
            if r == 'SUCCESS':
                return True
            else:
                raise error(r[0]['currentStatus']['commandStatus']['message'])
        else:
            r = r[0]['error']['errorMessage']['message']
            raise error(r)

    def ksql(self, ksql_string):
        r = self._request(endpoint='ksql', sql_string=ksql_string)

        if r.status_code == 200:
            r = r.json()
            return r
        else: 
            raise ValueError('Status Code: {}.\nMessage: {}'.format(r.status_code, r.content))

    def query(self, query_string, encoding='utf-8', chunk_size=128):
        """  
        Process streaming incoming data.

        """
        r = self._request(endpoint='query', sql_string=query_string)

        for chunk in r.iter_content(chunk_size=chunk_size):
            if chunk != b'\n':
                print(chunk.decode(encoding))

    def _request(self, endpoint, method='post', sql_string=''):
        url = '{}/{}'.format(self.url, endpoint)

        sql_string = self._validate_sql_string(sql_string) 
        data = json.dumps({
            "ksql": sql_string
        })
        
        headers = {
            "Content-Type": "application/json"
        }
        
        if endpoint == 'query':
            stream = True
        else:
            stream = False

        r = requests.request(
            method=method,
            url=url,
            data=data,
            timeout=self.timeout,
            headers=headers,
            stream=stream)  

        return r  

    @staticmethod
    def retry(exceptions, delay=1, max_retries=5):
        """
        A decorator for retrying a function call with a specified delay in case of a set of exceptions

        Parameter List
        -------------
        :param exceptions:  A tuple of all exceptions that need to be caught for retry
                                            e.g. retry(exception_list = (Timeout, Readtimeout))
        :param delay: Amount of delay (seconds) needed between successive retries.
        :param times: no of times the function should be retried

        """
        def outer_wrapper(function):
            @functools.wraps(function)
            def inner_wrapper(*args, **kwargs):
                final_excep = None  
                for counter in range(max_retries):
                    if counter > 0:
                        time.sleep(delay)
                    final_excep = None
                    try:
                        value = function(*args, **kwargs)
                        return value
                    except (exceptions) as e:
                        final_excep = e
                        pass #or log it

                if final_excep is not None:
                    raise final_excep
            return inner_wrapper
        return outer_wrapper


class SimplifiedAPI(BaseAPI):
    def __init__(self, url, **kwargs):
        super(SimplifiedAPI, self).__init__(url, **kwargs)

    def create_stream(self, table_name, columns_type, topic, value_format):
        return self._create(table_type='stream', 
                                    table_name = table_name, 
                                    columns_type = columns_type, 
                                    topic = topic, 
                                    value_format = value_format)

    def create_table(self, table_name, columns_type, topic, value_format):
        return self._create(table_type='table', 
                            table_name=table_name, 
                            columns_type=columns_type, 
                            topic=topic, 
                            value_format=value_format)

    def create_stream_as(self, table_name, select_columns, src_table, kafka_topic=None, 
              value_format='DELIMITED', conditions=[], partition_by=None, **kwargs):
        
        return self._create_as(table_type='stream',
                               table_name=table_name,
                               select_columns=select_columns,
                               src_table=src_table,
                               kafka_topic=kafka_topic,
                               value_format=value_format,
                               conditions=conditions,
                               partition_by=partition_by,
                               **kwargs)

    def _create(self, table_type, table_name, columns_type, topic, value_format):
        ksql_string = SQLBuilder.build(sql_type = 'create', 
                                      table_type = table_type, 
                                      table_name = table_name, 
                                      columns_type = columns_type, 
                                      topic = topic, 
                                      value_format = value_format)
        r = self.ksql(ksql_string)
        return self._parse_ksql_res(r, CreateError)

    @BaseAPI.retry(exceptions=(Timeout, CreateError))
    def _create_as(self, table_type, table_name, select_columns, src_table, kafka_topic=None, 
                   value_format='DELIMITED', conditions=[], partition_by=None, **kwargs):
        ksql_string = SQLBuilder.build(sql_type='create_as',
                                       table_type=table_type,
                                       table_name=table_name,
                                       select_columns=select_columns,
                                       src_table=src_table,
                                       kafka_topic=kafka_topic,
                                       value_format=value_format,
                                       conditions=conditions,
                                       partition_by=partition_by,
                                       **kwargs)
        r = self.ksql(ksql_string)
        return self._parse_ksql_res(r, CreateError)






__package_name__ = "ksql"
__ksql_server_version__ = "0.3"
__ksql_api_version__ = "0.1.2"
__version__ = __ksql_server_version__ + "." + __ksql_api_version__

from ksql.client import KSQLAPI
from ksql.builder import SQLBuilder
from ksql.api import SimplifiedAPI




class SQLTypeNotImplementYetError(Exception):

	def __init__(self, sql_type):
		self.msg = "This type {} has not be implement yet.".format(sql_type)

class IllegalTableTypeError(Exception):

	def __init__(self, table_type):
		self.msg = "This table type {} is illegal.".format(table_type)

class IllegalValueFormatError(Exception):

	def __init__(self, value_format):
		self.msg = "This value format {} is illegal.".format(value_format)


class CreateError(Exception):

	def __init__(self, e):
		self.msg = "{}".format(e)


class SQLFormatNotImplementError(Exception):
	pass

class BuildNotImplmentError(Exception):
	pass





from __future__ import absolute_import
from __future__ import print_function

import requests

from ksql.api import SimplifiedAPI


class KSQLAPI(object):
    """ API Class """

    def __init__(self, url, max_retries=3, **kwargs):
        self.url = url
        self.sa = SimplifiedAPI(url, max_retries=max_retries, **kwargs)

    def get_url(self):
        return self.url

    @property
    def timeout(self):
        return self.sa.get_timout()

    def get_ksql_version(self):
        r = requests.get(self.url)
        if r.status_code == 200:
            info = r.json().get('KSQL Server Info')
            version = info.get('version')
            return version
        else:
            raise ValueError('Status Code: {}.\nMessage: {}'.format(r.status_code, r.content))

    def ksql(self, ksql_string):
        return self.sa.ksql(ksql_string)
        
    def query(self, query_string, encoding='utf-8', chunk_size=128):
        self.sa.query(query_string=query_string, 
                       encoding=encoding, 
                       chunk_size=chunk_size)

    def create_stream(self, table_name, columns_type, topic, value_format):
        return self.sa.create_stream(table_name = table_name, 
                                    columns_type = columns_type, 
                                    topic = topic, 
                                    value_format = value_format)

    def create_table(self, table_name, columns_type, topic, value_format):
        return self.sa.create_table(table_name = table_name, 
                                    columns_type = columns_type, 
                                    topic = topic, 
                                    value_format = value_format)

    def create_stream_as(self, table_name, select_columns, src_table, kafka_topic=None, 
              value_format='DELIMITED', conditions=[], partition_by=None, **kwargs):
        return self.sa.create_stream_as(table_name=table_name,
                                        select_columns=select_columns,
                                        src_table=src_table,
                                        kafka_topic=kafka_topic,
                                        value_format=value_format,
                                        conditions=conditions,
                                        partition_by=partition_by,
                                        **kwargs)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
""" Setup module """
import os
import re

from setuptools import setup
from pip.req import parse_requirements

# Get version from __init__.py file
VERSION = "0.3.0.2.22"

here = os.path.dirname(__file__)

# Get long description
README = open(os.path.join(os.path.dirname(__file__), "README.rst")).read()

reqs = ['requests',
        'six',
        'urllib3']

# allow setup.py to be run from any path
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))

setup(
    name="ksql",
    version=VERSION,
    description="A Python wrapper for the KSql REST API",
    long_description=README,
    author="Bryan Yang @ Vpon",
    author_email="kenshin200528@gmail.com",
    url="https://github.com/bryanyang0528/ksql-python",
    license="MIT License",
    packages=[
        "ksql"
    ],
    include_package_data=True,
    platforms=['any'],
    install_requires=reqs,
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Natural Language :: English",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3.3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Topic :: Software Development :: Libraries :: Python Modules"
    ],
)
